"""Decorator-aware metadata extraction for magic specs."""

from __future__ import annotations

import ast
import copy
import inspect
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

from jaunt.registry import DecoratorApiRecord
from jaunt.spec_ref import SpecRef, spec_ref_from_object


@dataclass(frozen=True, slots=True)
class DecoratorAnalysis:
    auto_deps: tuple[SpecRef, ...]
    records: tuple[DecoratorApiRecord, ...]
    effective_signature: str | None
    effective_signature_source: Literal["decorated", "original"] | None
    warnings: tuple[str, ...]


def _find_nested_node(tree: ast.Module, qualname: str) -> ast.AST | None:
    parts = qualname.split(".")
    node: ast.AST = tree
    for part in parts:
        body = getattr(node, "body", None)
        if not isinstance(body, list):
            return None
        found = False
        for child in body:
            if (
                isinstance(child, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef))
                and child.name == part
            ):
                node = child
                found = True
                break
        if not found:
            return None
    return node


def _find_node_for_qualname(tree: ast.Module, qualname: str) -> ast.AST | None:
    if "." in qualname:
        return _find_nested_node(tree, qualname)
    for node in tree.body:
        if (
            isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef))
            and node.name == qualname
        ):
            return node
    return None


def _is_magic_decorator(node: ast.AST) -> bool:
    base = node.func if isinstance(node, ast.Call) else node

    if isinstance(base, ast.Name):
        return base.id == "magic"
    if isinstance(base, ast.Attribute):
        return (
            isinstance(base.value, ast.Name) and base.value.id == "jaunt" and base.attr == "magic"
        )
    return False


def _name_chain(node: ast.AST) -> list[str] | None:
    target = node.func if isinstance(node, ast.Call) else node
    if isinstance(target, ast.Name):
        return [target.id]
    if isinstance(target, ast.Attribute):
        parent = _name_chain(target.value)
        if not parent:
            return None
        return [*parent, target.attr]
    return None


def _symbol_paths_for_decorator(node: ast.AST) -> list[str]:
    chain = _name_chain(node)
    if not chain:
        return []
    out: list[str] = []
    for i in range(1, len(chain) + 1):
        out.append(".".join(chain[:i]))
    return out


def _resolve_path(module_name: str, symbol_path: str) -> object | None:
    mod = sys.modules.get(module_name)
    if mod is None:
        return None

    parts = [p for p in symbol_path.split(".") if p]
    if not parts:
        return None

    if not hasattr(mod, parts[0]):
        return None

    current: object = getattr(mod, parts[0])
    for part in parts[1:]:
        try:
            current = getattr(current, part)
        except Exception:
            return None
    return current


def _target_label(obj: object | None) -> str | None:
    if obj is None:
        return None
    module = getattr(obj, "__module__", None)
    qualname = getattr(obj, "__qualname__", None)
    if isinstance(module, str) and isinstance(qualname, str):
        return f"{module}.{qualname}"
    obj_t = type(obj)
    mod2 = getattr(obj_t, "__module__", None)
    q2 = getattr(obj_t, "__qualname__", None)
    if isinstance(mod2, str) and isinstance(q2, str):
        return f"{mod2}.{q2}"
    return None


def _best_effort_dep_ref(obj: object | None) -> SpecRef | None:
    if obj is None:
        return None
    try:
        return spec_ref_from_object(obj)
    except Exception:
        pass
    try:
        return spec_ref_from_object(type(obj))
    except Exception:
        return None


def _signature_quality(
    sig: inspect.Signature | None,
) -> Literal["good", "weak", "missing", "unknown"]:
    if sig is None:
        return "missing"
    params = list(sig.parameters.values())
    has_named = any(
        p.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        for p in params
    )
    annotated_params = any(p.annotation is not inspect.Signature.empty for p in params)
    has_ret = sig.return_annotation is not inspect.Signature.empty
    if has_named and (annotated_params or has_ret):
        return "good"
    return "weak"


def _signature_for_obj(
    obj: object | None,
) -> tuple[str | None, Literal["good", "weak", "missing", "unknown"]]:
    if obj is None:
        return None, "unknown"
    sig_target = obj if callable(obj) else type(obj)
    try:
        sig = inspect.signature(sig_target)
    except Exception:
        return None, "missing"
    return str(sig), _signature_quality(sig)


def _signature_from_function_node(node: ast.AST) -> str | None:
    if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
        return None

    # Keep this best-effort and representation-stable by asking AST unparse
    # for a normalized function header.
    clone = copy.deepcopy(node)
    clone.decorator_list = []
    clone.body = [ast.Pass()]
    try:
        rendered = ast.unparse(clone)
    except Exception:
        return None

    marker = (
        f"async def {node.name}" if isinstance(node, ast.AsyncFunctionDef) else f"def {node.name}"
    )
    i = rendered.find(marker)
    if i < 0:
        return None
    snippet = rendered[i:]
    header_lines: list[str] = []
    for line in snippet.splitlines():
        header_lines.append(line.strip())
        if line.rstrip().endswith(":"):
            break
    if not header_lines:
        return None
    header = " ".join(header_lines).rsplit(":", 1)[0]
    k = header.find("(")
    if k < 0:
        return None
    return header[k:]


def analyze_magic_decorators(
    *,
    module: str,
    qualname: str,
    source_file: str,
    decorated_obj: object,
) -> DecoratorAnalysis:
    try:
        source = Path(source_file).read_text(encoding="utf-8")
        tree = ast.parse(source, filename=source_file)
    except Exception:
        sig_text, _quality = _signature_for_obj(decorated_obj)
        return DecoratorAnalysis(
            auto_deps=(),
            records=(),
            effective_signature=sig_text,
            effective_signature_source="decorated" if sig_text is not None else None,
            warnings=(),
        )

    node = _find_node_for_qualname(tree, qualname)
    if node is None:
        sig_text, _quality = _signature_for_obj(decorated_obj)
        return DecoratorAnalysis(
            auto_deps=(),
            records=(),
            effective_signature=sig_text,
            effective_signature_source="decorated" if sig_text is not None else None,
            warnings=(),
        )

    decorators: list[ast.AST] = list(getattr(node, "decorator_list", []))
    magic_index: int | None = None
    for i, dec in enumerate(decorators):
        if _is_magic_decorator(dec):
            magic_index = i
            break

    dep_set: set[SpecRef] = set()
    records: list[DecoratorApiRecord] = []
    seen_records: set[tuple[str, str, str]] = set()

    for i, dec in enumerate(decorators):
        if _is_magic_decorator(dec):
            continue

        if magic_index is None:
            position: Literal["above_magic", "below_magic"] = "above_magic"
        else:
            position = "above_magic" if i < magic_index else "below_magic"

        expr = ast.get_source_segment(source, dec)
        if not isinstance(expr, str) or not expr.strip():
            try:
                expr = ast.unparse(dec)
            except Exception:
                expr = "<decorator>"

        for symbol_path in _symbol_paths_for_decorator(dec):
            dedupe_key = (symbol_path, position, expr)
            if dedupe_key in seen_records:
                continue
            seen_records.add(dedupe_key)

            resolved = _resolve_path(module, symbol_path)
            sig_text, quality = _signature_for_obj(resolved)
            record = DecoratorApiRecord(
                symbol_path=symbol_path,
                expression=expr,
                position=position,
                resolved_target=_target_label(resolved),
                signature=sig_text,
                annotation_quality=quality,
            )
            records.append(record)

            dep_ref = _best_effort_dep_ref(resolved)
            if dep_ref is not None:
                dep_set.add(dep_ref)

    warnings: list[str] = []

    dec_sig, dec_quality = _signature_for_obj(decorated_obj)
    src_sig = _signature_from_function_node(node)

    effective_sig: str | None = None
    effective_src: Literal["decorated", "original"] | None = None
    if dec_sig is not None and dec_quality == "good":
        effective_sig = dec_sig
        effective_src = "decorated"
    elif src_sig is not None:
        effective_sig = src_sig
        effective_src = "original"
        if dec_quality in ("weak", "missing"):
            warnings.append(
                f"weak decorator type metadata for {module}:{qualname}; "
                f"using source signature instead"
            )
    elif dec_sig is not None:
        effective_sig = dec_sig
        effective_src = "decorated"

    return DecoratorAnalysis(
        auto_deps=tuple(sorted(dep_set, key=lambda x: str(x))),
        records=tuple(records),
        effective_signature=effective_sig,
        effective_signature_source=effective_src,
        warnings=tuple(warnings),
    )


def resolve_qualname_for_line(*, source_file: str, line: int) -> tuple[str, str | None] | None:
    """Best-effort mapping of source location to function/class qualname."""
    try:
        source = Path(source_file).read_text(encoding="utf-8")
        tree = ast.parse(source, filename=source_file)
    except Exception:
        return None

    matches: list[tuple[int, int, str, str | None]] = []

    def walk(body: list[ast.stmt], prefix: str | None) -> None:
        for node in body:
            if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
                continue

            dec_lines = [
                d.lineno for d in getattr(node, "decorator_list", []) if hasattr(d, "lineno")
            ]
            start = min([node.lineno, *dec_lines])
            end = int(getattr(node, "end_lineno", node.lineno))
            if start <= line <= end:
                qual = f"{prefix}.{node.name}" if prefix else node.name
                class_name = (
                    prefix if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)) else None
                )
                depth = qual.count(".")
                span = end - start
                matches.append((depth, -span, qual, class_name))

            if isinstance(node, ast.ClassDef):
                next_prefix = f"{prefix}.{node.name}" if prefix else node.name
                walk(node.body, next_prefix)

    walk(tree.body, None)
    if not matches:
        return None
    matches.sort(reverse=True)
    _depth, _neg_span, qualname, class_name = matches[0]
    return qualname, class_name


_SIG_WARN_RE = re.compile(r"\*args|\*\*kwargs")


def signature_looks_variadic(sig_text: str | None) -> bool:
    if not sig_text:
        return False
    return bool(_SIG_WARN_RE.search(sig_text))



from __future__ import annotations

import asyncio
import logging
import os
from typing import Any

from jaunt.config import LLMConfig, PromptsConfig
from jaunt.errors import JauntConfigError
from jaunt.generate.base import GeneratorBackend, ModuleSpecContext, TokenUsage
from jaunt.generate.fingerprint import build_generation_fingerprint
from jaunt.generate.shared import (
    async_test_info,
    fmt_kv_block,
    load_prompt,
    render_template,
    strip_markdown_fences,
)

logger = logging.getLogger("jaunt.generate.anthropic")

_MAX_API_RETRIES = 4
_BASE_BACKOFF_S = 1.0


def _is_retryable(exc: BaseException) -> bool:
    """Return True for transient Anthropic API errors worth retrying."""
    cls_name = type(exc).__name__
    if cls_name in ("RateLimitError", "APITimeoutError", "APIConnectionError"):
        return True
    if cls_name == "APIStatusError":
        status = getattr(exc, "status_code", 0)
        return int(status) >= 500
    if isinstance(exc, (TimeoutError, ConnectionError, OSError)):
        return True
    return False


_ANTHROPIC_WRITE_MODULE_TOOL: dict[str, Any] = {
    "name": "write_module",
    "description": "Write the generated Python module source code.",
    "input_schema": {
        "type": "object",
        "properties": {
            "python_source": {
                "type": "string",
                "description": "The complete Python source code for the module.",
            },
            "imports_used": {
                "type": "array",
                "items": {"type": "string"},
                "description": "List of imported module names.",
            },
            "notes": {
                "type": "string",
                "description": "Optional generation notes.",
            },
        },
        "required": ["python_source"],
    },
}


class AnthropicBackend(GeneratorBackend):
    """Code generation backend using the Anthropic Messages API."""

    @property
    def supports_structured_output(self) -> bool:
        return True

    def __init__(self, llm: LLMConfig, prompts: PromptsConfig | None = None) -> None:
        api_key = (os.environ.get(llm.api_key_env) or "").strip()
        if not api_key:
            raise JauntConfigError(
                f"Missing API key: {llm.api_key_env}. "
                f"Set it in the environment or add it to <project_root>/.env."
            )
        self._model = llm.model
        self._thinking: dict[str, Any] | None = None
        if llm.anthropic_thinking_budget_tokens is not None:
            self._thinking = {
                "type": "enabled",
                "budget_tokens": llm.anthropic_thinking_budget_tokens,
            }

        try:
            from anthropic import AsyncAnthropic
        except ImportError as e:
            raise JauntConfigError(
                "The 'anthropic' package is required for provider='anthropic'. "
                "Install it with: pip install jaunt[anthropic]"
            ) from e

        self._client: Any = AsyncAnthropic(api_key=api_key)

        self._build_system = load_prompt(
            "build_system.md", prompts.build_system if prompts else None
        )
        self._build_module = load_prompt(
            "build_module.md", prompts.build_module if prompts else None
        )
        self._test_system = load_prompt("test_system.md", prompts.test_system if prompts else None)
        self._test_module = load_prompt("test_module.md", prompts.test_module if prompts else None)

    @property
    def provider_name(self) -> str:
        return "anthropic"

    def generation_fingerprint(self, ctx: ModuleSpecContext) -> str:
        prompt_parts = (
            [self._build_system, self._build_module]
            if ctx.kind == "build"
            else [self._test_system, self._test_module]
        )
        return build_generation_fingerprint(
            engine="legacy",
            kind=ctx.kind,
            prompt_parts=prompt_parts,
        )

    async def _call_anthropic(
        self, system: str, messages: list[dict[str, str]]
    ) -> tuple[str, TokenUsage | None]:
        """Call Anthropic Messages API with retry and exponential backoff."""
        last_exc: BaseException | None = None
        for attempt in range(_MAX_API_RETRIES):
            try:
                request_kwargs: dict[str, Any] = {
                    "model": self._model,
                    "max_tokens": 16384,
                    "system": system,
                    "messages": messages,
                }
                if self._thinking is not None:
                    request_kwargs["thinking"] = self._thinking
                resp: Any = await self._client.messages.create(**request_kwargs)
                content = resp.content
                if not content or not hasattr(content[0], "text"):
                    raise RuntimeError("Anthropic returned empty content.")
                usage = None
                if getattr(resp, "usage", None) is not None:
                    usage = TokenUsage(
                        prompt_tokens=getattr(resp.usage, "input_tokens", 0) or 0,
                        completion_tokens=getattr(resp.usage, "output_tokens", 0) or 0,
                        model=self._model,
                        provider="anthropic",
                    )
                return str(content[0].text), usage
            except Exception as exc:
                last_exc = exc
                if not _is_retryable(exc) or attempt >= _MAX_API_RETRIES - 1:
                    raise
                delay = _BASE_BACKOFF_S * (2**attempt)
                logger.warning(
                    "Anthropic API error (attempt %d/%d), retrying in %.1fs: %s",
                    attempt + 1,
                    _MAX_API_RETRIES,
                    delay,
                    exc,
                )
                await asyncio.sleep(delay)

        raise last_exc  # type: ignore[misc]

    async def _call_anthropic_structured(
        self, system: str, messages: list[dict[str, str]]
    ) -> tuple[str, TokenUsage | None]:
        """Call Anthropic Messages API with tool_use for structured output."""
        last_exc: BaseException | None = None
        for attempt in range(_MAX_API_RETRIES):
            try:
                request_kwargs: dict[str, Any] = {
                    "model": self._model,
                    "max_tokens": 16384,
                    "system": system,
                    "messages": messages,
                    "tools": [_ANTHROPIC_WRITE_MODULE_TOOL],
                    "tool_choice": {"type": "tool", "name": "write_module"},
                }
                if self._thinking is not None:
                    request_kwargs["thinking"] = self._thinking
                resp: Any = await self._client.messages.create(**request_kwargs)
                usage = None
                if getattr(resp, "usage", None) is not None:
                    usage = TokenUsage(
                        prompt_tokens=getattr(resp.usage, "input_tokens", 0) or 0,
                        completion_tokens=getattr(resp.usage, "output_tokens", 0) or 0,
                        model=self._model,
                        provider="anthropic",
                    )
                for block in resp.content:
                    if getattr(block, "type", None) == "tool_use" and block.name == "write_module":
                        source = block.input["python_source"]
                        imports = block.input.get("imports_used", [])
                        notes = block.input.get("notes", "")
                        if imports:
                            logger.debug("Structured output imports_used: %s", imports)
                        if notes:
                            logger.debug("Structured output notes: %s", notes)
                        return source, usage
                raise RuntimeError(
                    "Anthropic response did not contain a write_module tool_use block."
                )
            except Exception as exc:
                last_exc = exc
                if not _is_retryable(exc) or attempt >= _MAX_API_RETRIES - 1:
                    raise
                delay = _BASE_BACKOFF_S * (2**attempt)
                logger.warning(
                    "Anthropic API error (attempt %d/%d), retrying in %.1fs: %s",
                    attempt + 1,
                    _MAX_API_RETRIES,
                    delay,
                    exc,
                )
                await asyncio.sleep(delay)

        raise last_exc  # type: ignore[misc]

    def _render_messages(
        self, ctx: ModuleSpecContext, *, extra_error_context: list[str] | None
    ) -> tuple[str, list[dict[str, str]]]:
        """Return (system_prompt, messages) for the Anthropic Messages API."""
        expected = ", ".join(ctx.expected_names)

        spec_items: list[tuple[str, str]] = []
        for ref, source in sorted(ctx.spec_sources.items(), key=lambda kv: str(kv[0])):
            label = str(ref)
            prompt = ctx.decorator_prompts.get(ref)
            if prompt:
                source = f"{source.rstrip()}\n\n# Decorator prompt\n{prompt.rstrip()}\n"
            spec_items.append((label, source))

        deps_api_items: list[tuple[str, str]] = []
        for ref, api in sorted(ctx.dependency_apis.items(), key=lambda kv: str(kv[0])):
            deps_api_items.append((str(ref), api))

        deps_gen_items: list[tuple[str, str]] = []
        for mod, src in sorted(ctx.dependency_generated_modules.items(), key=lambda kv: kv[0]):
            deps_gen_items.append((mod, src))

        decorator_api_items: list[tuple[str, str]] = []
        for ref, api in sorted(ctx.decorator_apis.items(), key=lambda kv: str(kv[0])):
            decorator_api_items.append((str(ref), api))

        err_items: list[tuple[str, str]] = []
        if extra_error_context:
            for i, line in enumerate(extra_error_context, start=1):
                err_items.append((f"error_context[{i}]", line))

        mapping = {
            "spec_module": ctx.spec_module,
            "generated_module": ctx.generated_module,
            "expected_names": expected,
            "specs_block": fmt_kv_block(spec_items),
            "deps_api_block": fmt_kv_block(deps_api_items),
            "deps_generated_block": fmt_kv_block(deps_gen_items),
            "decorator_apis_block": fmt_kv_block(decorator_api_items),
            "module_contract_block": ctx.module_contract_block or "(none)\n",
            "blueprint_source_block": ctx.blueprint_source or "(none)\n",
            "attached_test_specs_block": ctx.attached_test_specs_block or "(none)\n",
            "package_context_block": ctx.package_context_block or "(none)\n",
            "error_context_block": fmt_kv_block(err_items),
            "async_test_info": async_test_info(ctx.async_runner),
        }

        if ctx.kind == "build":
            system_t = self._build_system
            user_t = self._build_module
        else:
            system_t = self._test_system
            user_t = self._test_module

        system = render_template(system_t, mapping).strip() + "\n"
        user = render_template(user_t, mapping).strip() + "\n"

        messages: list[dict[str, str]] = []
        if (ctx.skills_block or "").strip():
            skills_msg = "External library skills (reference):\n" + ctx.skills_block.strip() + "\n"
            messages.append({"role": "user", "content": skills_msg})
            messages.append(
                {"role": "assistant", "content": "Understood. I'll reference these libraries."}
            )
        messages.append({"role": "user", "content": user})
        return system, messages

    async def generate_module(
        self,
        ctx: ModuleSpecContext,
        *,
        extra_error_context: list[str] | None = None,
    ) -> tuple[str, TokenUsage | None]:
        system, messages = self._render_messages(ctx, extra_error_context=extra_error_context)
        if self.supports_structured_output:
            return await self._call_anthropic_structured(system, messages)
        raw, usage = await self._call_anthropic(system, messages)
        return strip_markdown_fences(raw), usage



"""Pure helpers for mapping spec modules to generated modules and file paths."""

from __future__ import annotations

from pathlib import Path


def spec_module_to_generated_module(module: str, generated_dir: str = "__generated__") -> str:
    parts = module.split(".")
    if len(parts) == 1:
        return f"{parts[0]}.{generated_dir}"
    if len(parts) >= 2 and parts[1] == generated_dir:
        return module
    return ".".join([parts[0], generated_dir, *parts[1:]])


def module_to_relpath(module: str) -> Path:
    parts = module.split(".")
    if len(parts) == 1:
        return Path(parts[0]) / "__init__.py"
    parts = [*parts[:-1], f"{parts[-1]}.py"]
    return Path(*parts)


def generated_module_to_relpath(module: str, *, generated_dir: str = "__generated__") -> Path:
    parts = module.split(".")
    if parts and parts[-1] == generated_dir:
        return Path(*parts) / "__init__.py"
    return module_to_relpath(module)



"""Vigil in the forests of the night -- watch mode, rebuild on spec changes."""

from __future__ import annotations

import argparse
import inspect
import time
from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine
from dataclasses import dataclass
from pathlib import Path
from typing import Any


@dataclass(frozen=True, slots=True)
class WatchEvent:
    """A batch of relevant file changes."""

    changed_paths: frozenset[Path]
    timestamp: float


@dataclass(frozen=True, slots=True)
class WatchCycleResult:
    """Result of a single watch rebuild cycle."""

    build_exit_code: int
    test_exit_code: int | None
    duration_s: float
    changed_paths: frozenset[Path]


def check_watchfiles_available() -> None:
    """Raise ImportError with a helpful message if watchfiles is not installed."""
    import importlib

    try:
        importlib.import_module("watchfiles")
    except ImportError:
        raise ImportError(
            "watchfiles is required for watch mode. Install it with: pip install jaunt[watch]"
        ) from None


def filter_spec_files(
    changed_paths: frozenset[Path],
    *,
    source_roots: list[Path],
    test_roots: list[Path],
    generated_dir: str = "__generated__",
) -> frozenset[Path]:
    """Filter changed paths to .py files under source/test roots, excluding __generated__."""
    roots = [*source_roots, *test_roots]
    kept: set[Path] = set()
    for p in changed_paths:
        if p.suffix != ".py":
            continue
        if generated_dir and generated_dir in p.parts:
            continue
        if any(p.is_relative_to(r) for r in roots):
            kept.add(p)
    return frozenset(kept)


async def run_watch_loop(
    *,
    changes_iter: AsyncIterator[set[tuple[Any, str]]],
    run_cycle: Callable[[WatchEvent], WatchCycleResult | Awaitable[WatchCycleResult]],
    on_event: Callable[[str], None],
    on_cycle_result: Callable[[WatchCycleResult], None],
    on_error: Callable[[BaseException], None],
    source_roots: list[Path],
    test_roots: list[Path],
    generated_dir: str = "__generated__",
) -> None:
    """Main watch loop. Consumes changes_iter, filters, and calls run_cycle."""
    async for raw_changes in changes_iter:
        paths = frozenset(Path(p) for _, p in raw_changes)
        relevant = filter_spec_files(
            paths,
            source_roots=source_roots,
            test_roots=test_roots,
            generated_dir=generated_dir,
        )
        if not relevant:
            continue

        event = WatchEvent(changed_paths=relevant, timestamp=time.monotonic())

        names = ", ".join(str(p) for p in sorted(relevant))
        on_event(f"[watch] change detected: {names}")
        on_event("[watch] building...")

        try:
            result = run_cycle(event)
            if inspect.isawaitable(result):
                result = await result
        except Exception as exc:
            on_error(exc)
            continue

        on_event(f"[watch] done ({result.duration_s:.1f}s)")
        on_cycle_result(result)


def format_watch_cycle_json(result: WatchCycleResult) -> dict[str, object]:
    """Format a cycle result as a JSON-serializable dict."""
    ok = result.build_exit_code == 0 and result.test_exit_code in (None, 0)
    return {
        "command": "watch",
        "ok": ok,
        "build_exit_code": result.build_exit_code,
        "test_exit_code": result.test_exit_code,
        "duration_s": round(result.duration_s, 2),
        "changed_paths": sorted(str(p) for p in result.changed_paths),
    }


def build_cycle_runner(
    args: Any,
    *,
    run_tests: bool,
) -> Callable[[WatchEvent], Coroutine[Any, Any, WatchCycleResult]]:
    """Create a cycle runner that calls async build/test helpers."""
    from jaunt.cli import _cmd_build_async, _cmd_test_async

    build_args = argparse.Namespace(
        root=getattr(args, "root", None),
        config=getattr(args, "config", None),
        jobs=getattr(args, "jobs", None),
        force=bool(getattr(args, "force", False)),
        target=list(getattr(args, "target", [])),
        no_infer_deps=bool(getattr(args, "no_infer_deps", False)),
        no_progress=bool(getattr(args, "no_progress", False)),
        no_cache=bool(getattr(args, "no_cache", False)),
        json_output=False,
    )

    test_args = argparse.Namespace(
        root=getattr(args, "root", None),
        config=getattr(args, "config", None),
        jobs=getattr(args, "jobs", None),
        force=bool(getattr(args, "force", False)),
        target=list(getattr(args, "target", [])),
        no_infer_deps=bool(getattr(args, "no_infer_deps", False)),
        no_progress=bool(getattr(args, "no_progress", False)),
        no_cache=bool(getattr(args, "no_cache", False)),
        json_output=False,
        no_build=True,
        no_run=False,
        pytest_args=[],
    )

    async def runner(event: WatchEvent) -> WatchCycleResult:
        t0 = time.monotonic()
        build_rc = await _cmd_build_async(build_args)

        test_rc: int | None = None
        if run_tests and build_rc == 0:
            test_rc = await _cmd_test_async(test_args)

        duration = time.monotonic() - t0
        return WatchCycleResult(
            build_exit_code=build_rc,
            test_exit_code=test_rc,
            duration_s=duration,
            changed_paths=event.changed_paths,
        )

    return runner


def make_watchfiles_iter(
    watch_paths: list[Path],
) -> AsyncIterator[set[tuple[Any, str]]]:
    """Create an async iterator using watchfiles.awatch()."""
    import watchfiles

    return watchfiles.awatch(*watch_paths, debounce=200)
